pub const KERNELS_SRC: &str = concat!(
    include_str!("kernels/part1.cu"),
    include_str!("kernels/part2.cu"),
    include_str!("kernels/part3.cu"),
    include_str!("kernels/part4.cu"),
    include_str!("kernels/part5.cu"),
    include_str!("kernels/part6.cu"),
);
